
(**
 * Memory leak detector for Virtual Pascal and Delphi.
 *
 * (c) 1998 Joerg Pleumann <pleumann@uni-duisburg.de>
 *
 * Last changed 30-May-98, comments appreciated.
 *
 * Added location info 15-Jan-98,
 * Antony T Curtis <antony.curtis@olcs.net>
 * 1999.01.15 * Veit Kannegieser * more portable
 * 1998.01.15 * Veit Kannegieser * get caller of GetMem/ReAllocMem
 *
 * This unit implements and installs a new memory manager
 * that calls the GetMem / FreeMem / ReallocMem functions
 * of the original one (see TMemoryManager in System unit),
 * but performs some additional heap checking: All memory
 * allocations are stored in a list and removed when a
 * corresponding deallocation occurs. If, at program exit,
 * this list is non-empty, the program does two things:
 *
 * (1) It produces three beeps to inform the user or
 *     the developer about memory leaks.
 *
 * (2) It writes a dump of all leaks and their contents
 *     to a file "MemLeaks.dmp" in the current directory.
 *     The contents of a memory leak usually give you a
 *     good hint where to look for the leak. The unit
 *     also tries to categorize each memory leak by
 *     looking at its contents. Normally, AnsiStrings
 *     and class instances (new object model) are
 *     detected properly.
 *
 * Leak detection is only active when ExitCode (see System
 * unit) is zero, which indicates normal program termination.
 * A non-zero ExitCode usually means abnormal termination due
 * to an error. It makes no sense to dump the heap in that
 * case, since the application usually doesn't get any chance
 * to clean up the heap if a run-time error occurs.
 *
 * How to use the unit? Quite easy: Simply add it to the
 * "uses" list of your main executable, preferably as last
 * entry. That's it. Note that your program will run a
 * little slower with leak detection included, so you might
 * want to use this unit only while developing / debugging.
 *
 * Have fun!
 *)

unit MemLeaks;

{$H-}
{$X+,Delphi+} (* uses "Result" keyword *)

interface

uses
  SysUtils;

implementation

{$IFNDEF VIRTUALPASCAL} (* DELPHI ? *)
uses
  Windows;
{$ENDIF}

{$IFDEF VIRTUALPASCAL}
uses
  VpSysLow;
{$ENDIF VIRTUALPASCAL}

(**
 * Forwards for the three functions implementing
 * the new memory manager.
 *)

function NewGetMem(Size: LongInt): Pointer; forward;
function NewFreeMem(P: Pointer): LongInt; forward;
function NewReallocMem(P: Pointer; Size: LongInt): Pointer; forward;

const

  (**
   * Growth of memory block list.
   *)

  MemBlockDelta = 1024;

  (**
   * The new memory manager. The three functions
   * variables point to the new memory management
   * functions.
   *)

  NewMemoryManager: TMemoryManager =
  ( GetMem:     NewGetMem;
    FreeMem:    NewFreeMem;
    ReAllocMem: NewReAllocMem );

type

  (**
   * An entry of the memory block list.
   *)

  PMemBlock = ^TMemBlock;
  TMemBlock = record
    FAddress: Pointer;
    FSize:    LongInt;
    FCaller:  Pointer;  (* ATC -- Stores the caller's address *)
    FThread:  LongInt;  (* ATC -- Stores the caller's thread  *)
  end;

  (*
   * The list of memory blocks.
   *)

  PMemBlockList = ^TMemBlockList;
  TMemBlockList = array[0..MaxInt div SizeOf(TMemBlock) - 1] of TMemBlock;

var

  (**
   * Points to the list of currently
   * allocated memory blocks. *)

  MemBlockList:  PMemBlockList = nil;

  (**
   * Holds the number of currently
   * allocated memory blocks.
   *)

  MemBlockCount: LongInt = 0;

  (**
   * Holds the current size of the
   * memory block list.
   *)

  MemBlockLimit: LongInt = 0;

  (**
   * Holds the total amount of memory
   * allocated by the program, not
   * counting the overhead needed for
   * heap management.
   *)

  MemBlockTotal: LongInt = 0;

  (**
   * Holds the memory manager that was
   * active before the memory leak
   * detector.
   *)

  OldMemoryManager: TMemoryManager;

(**
 * ATC -- Finds out the address of the caller
 * with a quick walk through the stack
 *)

function CallerAddr2org:Pointer; {$FRAME-}
assembler;
asm
  mov   eax, [ebp]
  mov   eax, [eax+4]
  sub   eax, 5
end;

function CallerAddr2:Pointer;
(*$FRAME-*)(*$USES NONE*)
  asm
    mov eax,[ebp+4]
    sub eax,5

    (*$IFDEF VIRTUALPASCAL*) (* VP 2.0 specific *)
    // inside "_MemNew" ?
    mov edx,offset _MemNew
    cmp eax,edx
    jb @not_MemNew
    add edx,$20
    cmp eax,edx
    jnb @not_MemNew

    // Who has called "New" or "GetMem" ?
    mov eax,[ebp+$14]
    sub eax,5
    jmp @ret

  @not_MemNew:

    // inside "_MemRealloc" (Realloc) ?
    mov edx,offset _MemRealloc
    cmp eax,edx
    jb @not_MemRealloc
    add edx,$20
    cmp eax,edx
    jnb @not_MemRealloc_0
    add edx,$56-$20
    cmp eax,edx
    jnb @not_MemRealloc

    // Who has called "ReallocMem" ?
    mov eax,[ebp+$1c]
    sub eax,5
    jmp @ret

  @not_MemRealloc_0:
    mov eax,[ebp+$18]
    sub eax,5
    jmp @ret

  @not_MemRealloc:

    // more detection ..

    (*$ENDIF*)
  @ret:
  end;

(**
 * Adds a block to the list of currrently
 * allocated memory blocks.
 *)

procedure AddBlock(Address: Pointer; Size: LongInt; Caller: Pointer);
begin
  if Address <> nil then
  begin
    if MemBlockCount = MemBlockLimit then
    begin
      MemBlockList := OldMemoryManager.ReAllocMem(MemBlockList,
        (MemBlockLimit + MemBlockDelta) * SizeOf(TMemBlock));

      Inc(MemBlockLimit, MemBlockDelta);
    end;

    with MemBlockList^[MemBlockCount] do
    begin
      FAddress := Address;
      FSize := Size;
      FCaller := Caller;
      FThread := GetThreadID;
    end;

    Inc(MemBlockCount);
    Inc(MemBlockTotal, Size);
  end;
end;

(**
 * Deletes a block from the list of currrently
 * allocated memory blocks.
 *)

procedure DeleteBlock(Address: Pointer);
var
  I: LongInt;
begin
  if Address <> nil then
  begin
    I := MemBlockCount - 1;

    while (I <> -1) and (MemBlockList^[I].FAddress <> Address) do
      Dec(I);

    if I <> - 1 then
    begin
      Dec(MemBlockCount);
      Dec(MemBlockTotal, MemBlockList^[I].FSize);

      Move(MemBlockList^[I + 1], MemBlockList^[I],
        (MemBlockCount - I) * SizeOf(TMemBlock));
    end;
  end;
end;

(**
 * Tries to classify the memory block. This is more or
 * less a guess, but works most of the time. :-) The
 * function is able to detect AnsiStrings and object
 * instances (new object model only). Everything else
 * is returned as 'Unknown'.
 *)

function MemClassify(Address: Pointer; Size: LongInt): string;
type
  TStrRec = record
    RefCnt: LongInt;
    Length: LongInt;
    Data:   array[0..MaxLongInt] of Char;
  end;
var
  AString: ^TStrRec absolute Address;
  AObject: TObject absolute Address;
begin
  if Size > 8 then
    with AString^ do
    begin
      if (Length + 9 = Size) and (Data[Length] = #0) then
      begin
        Result := 'AnsiString (Length=' + IntToStr(Length)
                + ', RefCnt=' + IntToStr(RefCnt) + ')';
        Exit;
      end;
    end;

  try
    if AObject.InstanceSize = Size then
    begin
      Result := TObject(Address).ClassName;
      Exit;
    end;
  except
  end;

  Result := 'Unknown';
end;

(**
 * Dumps a given block of memory to a string.
 *)

function MemToHexAsc(Address: PChar; Length: Integer): string;
var
  I: Integer;
begin
  Result := '';

  for I := 0 to Length - 1 do
    Result := Result + IntToHex(Byte(Address[I]), 2) + ' ';

  for I := Length to 15 do
    Result := Result + '   ';

  Result := Result + '   ';

  for I := 0 to Length - 1 do
  begin
    if Address[I] >= ' ' then
      Result := Result + Address[I]
    else
      Result := Result + '.';
  end;
end;

function CallerInfo(Caller: Pointer; Thread:LongInt): string;
var
  LineNum               : Integer;
begin
  GetLocationInfo(Caller,Result,LineNum);
  Result := Format('%s line %d (thread %d)',[Result,LineNum,Thread]);
end;

(**
 * Produces a HEX- / ASCII dump of the memory block
 * list into a file called "MemLeaks.dmp".
 *)

procedure DumpHeap;
var
  F: Text;
  I, L: LongInt;
  P: PChar;
begin
  Assign(F, 'MemLeaks.dmp');
  Rewrite(F);

  WriteLn(F, '; Dump of memory leak detector.');
  WriteLn(F, ';');
  WriteLn(F, '; Executable file  : ', ParamStr(0));
  WriteLn(F, '; Date and time    : ', FormatDateTime('yyyy"/"mm"/"dd hh":"nn":"ss', Now));
  WriteLn(F, '; Number of leaks  : ', MemBlockCount);
  WriteLn(F, '; Total memory loss: ', MemBlockTotal, ' bytes (without overhead)');
  WriteLn(F, ';');
  WriteLn(F, '; Leaks are listed in order of allocation.');
  WriteLn(F);

  for I := 0 to MemBlockCount - 1 do
  begin
    with MemBlockList^[I] do
    begin
      P := FAddress;
      L := FSize;

      WriteLn(F, 'Address : ', IntToHex(LongInt(P), 8));
      WriteLn(F, 'Size    : ', IntToHex(L, 8), ' (', L, ' bytes)');
      WriteLn(F, 'Caller  : ', CallerInfo(FCaller, FThread));
      WriteLn(F, 'Type    : ', MemClassify(FAddress, FSize));
      Write  (F, 'Contents: ');

      while L > 16 do
      begin
        WriteLn(F, MemToHexAsc(P, 16));
        Write  (F, '          ');
        Inc(P, 16);
        Dec(L, 16);
      end;

      WriteLn(F, MemToHexAsc(P, L));
      WriteLn(F);
    end;
  end;

  WriteLn(F, '; End of file.');

  Close(F);
end;

(**
 * Implements the GetMem function of the new memory
 * manager. Simply calls the original memory manager
 * and then then adds the new block to the list.
 *)

function NewGetMem(Size: LongInt): Pointer;
begin
  Result := OldMemoryManager.GetMem(Size);
  AddBlock(Result, Size, CallerAddr2);
end;

(**
 * Implements the FreeMem function of the new memory
 * manager. Simply removes the block from the list
 * and then calls the original memory manager.
 *)

function NewFreeMem(P: Pointer): Longint;
begin
  DeleteBlock(P);
  Result := OldMemoryManager.FreeMem(P);
end;

(**
 * Implements the ReallocMem function of the new memory
 * manager. Removes the old block from the list, calls
 * the original memory manager and adds the new block to
 * the list.
 *)

function NewReallocMem(P: Pointer; Size: LongInt): Pointer;
begin
  DeleteBlock(P);
  Result := OldMemoryManager.ReAllocMem(P, Size);
  AddBlock(Result, Size, CallerAddr2);
end;

(**
 * The initialization part reserves some initial space for
 * the memory block list and installs the new memory
 * manager.
 *)

initialization
  GetMem(MemBlockList, MemBlockDelta * SizeOf(TMemBlock));

  GetMemoryManager(OldMemoryManager);
  SetMemoryManager(NewMemoryManager);

(**
 * The finalization part restores the original memory manager.
 * Afterwards, it checks whether there are still entries left
 * in the memory block list. If so, it beeps three times and
 * dumps all the memory leaks to a file.
 *)

finalization
  SetMemoryManager(OldMemoryManager);

  if (ExitCode = 0) and (MemBlockCount <> 0) then
  begin
    {$IFNDEF VIRTUALPASCAL} (* DELPHI ? *)
    MessageBeep($FFFFFFFF);
    MessageBeep($FFFFFFFF);
    MessageBeep($FFFFFFFF);
    {$ENDIF}

    {$IFDEF VIRTUALPASCAL} (* VP 2.0 OS/2,Win32,DPMI32 *)
    SysBeepEx(440, 100);
    SysCtrlSleep(500);
    SysBeepEx(440, 100);
    SysCtrlSleep(500);
    SysBeepEx(440, 100);
    {$ENDIF VIRTUALPASCAL}

    DumpHeap;
  end;

  FreeMem(MemBlockList);
end.

